from coalib.bearlib.aspects import Taste, Root


@Root.subaspect
class Smell:
    """
    This aspect detects `code smells` or `bad smells` in your code.

    `Smells` are certain structures in a code that indicate violation of
    fundamental design principles. They are usually not bugs; they are not
    technically incorrect and do not currently prevent the program from
    functioning.
    """
    class docs:
        example = """
        * Feature envy
        * Data clump
        * Too large class
        * Too long parameter list
        etc...
        """
        example_language = 'English'
        importance_reason = """
        Even though they are not necessarily bugs, code smells increase the
        risk of bugs or failure in the future and may slow down development.
        """
        fix_suggestions = """
        There are several `refactoring techniques` that can be used to deal
        with `code smells` including:

        * Composing methods
        * Moving features between objects
        * Organizing data
        * Simplifying conditional expressions
        * Simplifying method calls
        * Dealing with generalisation

        See <https://sourcemaking.com/refactoring/refactorings> for more
        information.
        """


@Smell.subaspect
class ClassSmell:
    """
    This aspect detects `code smells` or `bad smells` related to classes'
    definitions in your codebase.

    Class-level code smells indicate poorly defined classes (including too
    large classes or God object, data clump feature envy etc...) in your
    source code.
    """
    class docs:
        example = """
        * Too large classes
        * Data clump
        * Feature envy
        etc ...
        """
        example_language = 'English'
        importance_reason = """
        These classes should be refactored for better readability and
        maintainability of your source code.
        """
        fix_suggestions = """
        When a class is wearing too many (functional) hats (too large
        classes), you should probably think about splitting it up:

        * Extract class
        * Extract subclass
        * Extract interface
        """


@Smell.subaspect
class MethodSmell:
    """
    This aspect detects `code smells` related to methods' and functions
    definitions in your codebase.

    Method-level code smells indicate poorly defined method and or
    functions (too long method or functions, or functions with too many
    parameters) in your source code.
    """
    class docs:
        example = """
        # This function has way too many parameters

        def func(a, b, c, d, e, f, g, h, i, j, k, l, m, n, o, p, q, r, s, t, z):
            pass
        """
        example_language = 'python'
        importance_reason = """
        Make your functions and methods unambiguous, easy to read and debug
        by reducing the number of parameters and length of your methods and
        functions.
        """
        fix_suggestions = """
        A fix for this would simply consist of redefining the functions
        (and or method), making them shorter and reducing the number of
        parameters (maybe by creating more functions or using libraries).
        """

    max_method_length = Taste[int](
        'Represents the max number of lines for a method or a function\'s'
        'definition.',
        (40,), default=40)
    max_parameters = Taste[int](
        'Represents the max number of parameters for a function\'s.',
        (5, 10), default=10)


@ClassSmell.subaspect
class DataClump:
    """
    This aspect detects `data clumps` in you codebase.

    `Data clump` is a name given to any group of variables which are passed
    around together (in a clump) throughout various parts of the program.
    """
    class docs:
        example = """
        public static void main(String args[]) {

            String firstName = args[0];
            String lastName = args[1];
            Integer age = new Integer(args[2]);
            String gender = args[3];
            String occupation = args[4];
            String city = args[5];

            welcomeNew(firstName,lastName,age,gender,occupation,city);
        }

        public static void welcomeNew(String firstName, String lastName,
                              Integer age, String gender,
                              String occupation, String city){

            System.out.printf("Welcome %s %s, a %d-year-old %s "\
"from %s who works as a%s\n",firstName, lastName, age, gender, city, \
occupation);
        }
        """
        example_language = 'java'
        importance_reason = """
        Data clumps make codes difficult to read, debug, scale, and also
        hardly reusable.
        """
        fix_suggestions = """
        Formally group the different variables together into a single object.
        """


@ClassSmell.subaspect
class ClassLength:
    """
    This aspect checks on classes' definitions length in your codebase.
    """
    class docs:
        example = """
        // This is large class given that the `max_class_length` is 20

        public class Employee
        {
            private float salary;
            private float bonusPercentage;
            private EmployeeType employeeType;
            public Employee(float salary, float bonusPercentage,
                            EmployeeType employeeType)
            {
                this.salary = salary;
                this.bonusPercentage = bonusPercentage;
                this.employeeType = employeeType;
            }
            public float CalculateSalary()
            {
                switch (employeeType)
                {
                    case EmployeeType.Worker:
                        return salary;
                    case EmployeeType.Supervisor:
                        return salary + (bonusPercentage * 0.5F);
                    case EmployeeType.Manager:
                        return salary + (bonusPercentage * 0.7F);
                }
                return 0.0F;
            }
        }
        """
        example_language = 'Java'
        importance_reason = """
        Too large classes also known as God objects, result in ambiguous and
        difficult to debug source code; whereas too small classes (or lazy
        classes or freeloader) are sometimes useless.
        """
        fix_suggestions = """
        Usually splitting up those classes into many other classes solves the
        problem.
        """
    max_class_length = Taste[int](
        'Represents the max number of lines for a class\'s definition.',
        (999,), default=999)

    # This shall probably be removed.
    min_class_length = Taste[int](
        'Represents the min number of lines for a class\'s definition.',
        (1,), default=1)


@ClassSmell.subaspect
class FeatureEnvy:
    """
    This aspect detects occurrences of feature envy in your codebase.

    `Feature envy` describes classes that excessively use methods of other
    classes.
    """
    class docs:
        example = """
        public class Phone {
            private final String unformattedNumber;
            public Phone(String unformattedNumber) {
                this.unformattedNumber = unformattedNumber;
            }
            public String getAreaCode() {
                return unformattedNumber.substring(0,3);
            }
            public String getPrefix() {
                return unformattedNumber.substring(3,6);
            }
            public String getNumber() {
                return unformattedNumber.substring(6,10);
            }
        }
        public class Customer…
            private Phone mobilePhone;
            public String getMobilePhoneNumber() {
                return "(" +
                    mobilePhone.getAreaCode() + ") " +
                    mobilePhone.getPrefix() + "-" +
                    mobilePhone.getNumber();
            }
        }
        """
        example_language = 'java'
        importance_reason = """
        This smell may occur after fields are moved to a data class; which
        makes the code less readable, and difficult to debug.
        """
        fix_suggestions = """
        Move the operations on data to the newly defined class(given that
        the fields of one class were moved to this class) as well.
        """


@Smell.subaspect
class Naming:
    """
    This aspect checks on identifiers in your codebase (their length
    and the appropriate naming convention to use for them, be it variables,
    classes or functions names.)
    """
    class docs:
        example = """
        camelCase naming convention, snake_case naming convention,
        hyphenated-case naming convention etc...
        """
        example_language = 'English'
        importance_reason = """
        Consistent use of naming convention, make the code easy to read
        and debug.
        """
        fix_suggestions = """
        Use the appropriate naming convention for each data type.
        """

    variable_naming_convention = Taste[str](
        'Naming convention to use for variables\'s identifiers',
        ('lowerCamelCase', 'snake_case', 'kebab-case', 'UpperCamelCase'),
        default='snake_case')
    function_naming_convention = Taste[str](
        'Naming convention to use for functions\'s or methods\'s identifiers',
        ('lowerCamelCase', 'snake_case', 'kebab-case', 'UpperCamelcase'),
        default='snake_case')
    class_naming_convention = Taste[str](
        'Naming convention to use for classes\'s identifiers',
        ('lowerCamelCase', 'snake_case', 'kebab-case', 'UpperCamelCase'),
        default='UpperCamelCase')
    max_identifier_length = Taste[int](
        'The maximum number of character for an identifier.',
        (31,), default=31)


@Smell.subaspect
class Complexity:
    """
    This aspect checks on the cyclomatic complexity of your code.
    """
    class docs:
        example = """
        for (i=0; i<n; ++i){
            for (i=0; i<n; ++i){
                for (i=0; i<n; ++i){
                    for (i=0; i<n; ++i){
                        for (i=0; i<n; ++i){
                            for (i=0; i<n; ++i){
                                for (i=0; i<n; ++i){
                                    for (i=0; i<n; ++i){
                                        ...
                                        //do something
        ...
        }
        """
        example_language = 'C++'
        importance_reason = """
        Very complex code are difficult to read, debug and maintain.
        It is always a good idea to keep things as simple as possible.
        """
        fix_suggestions = """
        This can be solved by breaking down complex functions into smaller
        onces.
        """
    cyclomatic_complexity = Taste[int](
        'This the maximum number of embedded branches or embedded loops'
        ' allowed.',
        (6,), default=6)



from coalib.bearlib.aspects import Root


@Root.subaspect
class Spelling:
    """
    This aspect describes spelling of your source code.
    """
    class docs:
        example = """
        `Coala`, `AspectsYEAH`
        """
        example_language = 'All'
        importance_reason = """
        They are words we always want to be written the way we want like the
        name of our app eg: `coala`...
        """
        fix_suggestions = """
        Use coala :) to detect what word is not written as expected, given
        that there is a bear for it ;).
        """


@Spelling.subaspect
class aspectsYEAH:
    """
    This aspect dictates that the term ``aspects`` and ``aspect`` must have
    all letters in lower case and the term ``aspectsYEAH`` must match the
    exact expression.
    """
    class docs:
        example = """
        Valid Cases:
        # aspects are the mother of all futures of coala
        # ``aspectsYEAH`` project is simply awesome
        Invalid Cases:
        # Aspects are the mother of all futures of coala
        # ``aspectsyeah`` project is simply awesome
        """
        example_language = 'All'
        importance_reason = """
        The concepts of aspects in coala are too fundamental to contain any
        upper-case letters.
        """
        fix_suggestions = """
        ``aspects`` or ``aspect`` are always written with all lower-case
        letters or the term ``aspectsYEAH`` must match the exact expression.
        """


@Spelling.subaspect
class coalaCorrect:
    """
    This aspect dictates that the term ``coala`` must always be written with
    with a lowercase `c`.
    """
    class docs:
        example = """
        Valid Cases:
        # ``coala`` is so much fun!
        # ``coalA`` hmm... not exactly ``coala`` but it is fine ;)
        Invalid Cases:
        # ``Coala`` hmm... this is wrong :(
        # ``CoalA``... @cobot please do something about this :(
        """
        example_language = 'All'
        importance_reason = """
        We like it this way ;).
        """
        fix_suggestions = """
        Always write ``coala`` with a lower case ``c``.
        """



from collections import namedtuple

from coala_utils.decorators import generate_eq, generate_repr
from coalib.results.TextRange import TextRange
from functools import lru_cache


@generate_repr()
@generate_eq('documentation', 'language', 'docstyle',
             'indent', 'marker', 'position')
class DocumentationComment:
    """
    The DocumentationComment holds information about a documentation comment
    inside source-code, like position etc.
    """
    Parameter = namedtuple('Parameter', 'name, desc')
    ExceptionValue = namedtuple('ExceptionValue', 'name, desc')
    ReturnValue = namedtuple('ReturnValue', 'desc')
    Description = namedtuple('Description', 'desc')

    def __init__(self, documentation, docstyle_definition,
                 indent, marker, position):
        """
        Instantiates a new DocumentationComment.

        :param documentation:
            The documentation text.
        :param docstyle_definition:
            The ``DocstyleDefinition`` instance that defines what docstyle is
            being used in the documentation.
        :param indent:
            The string of indentation used in front of the first marker of the
            documentation.
        :param marker:
            The three-element tuple with marker strings, that identified this
            documentation comment.
        :param position:
            The starting ``TextPosition`` of the documentation.
        """
        self.documentation = documentation
        self.docstyle_definition = docstyle_definition
        self.indent = '' if indent is None else indent
        self.marker = ('', '', '') if marker is None else marker
        self.position = position
        self.range = None if position is None else TextRange.from_values(
            position.line,
            position.column,
            position.line + self.assemble().count('\n'),
            len(self.assemble()) - self.assemble().rfind('\n'))

    def __str__(self):
        return self.documentation

    @property
    def language(self):
        return self.docstyle_definition.language

    @property
    def docstyle(self):
        return self.docstyle_definition.docstyle

    @property
    def metadata(self):
        return self.docstyle_definition.metadata

    def parse(self):
        """
        Parses documentation independent of language and docstyle.

        :return:
            The list of all the parsed sections of the documentation. Every
            section is a namedtuple of either ``Description`` or ``Parameter``
            or ``ReturnValue``.
        :raises NotImplementedError:
            When no parsing method is present for the given language and
            docstyle.
        """
        if self.language == 'python' and self.docstyle == 'default':
            return self._parse_documentation_with_symbols(
                (':param ', ':'), (':raises ', ':'), ':return:')
        elif self.language == 'python' and self.docstyle == 'doxygen':
            return self._parse_documentation_with_symbols(
                ('@param ', ' '), ('@raises ', ' '), '@return ')
        elif self.language == 'java' and self.docstyle == 'default':
            return self._parse_documentation_with_symbols(
                ('@param  ', ' '), ('@raises  ', ' '), '@return ')
        elif self.language == 'golang' and self.docstyle == 'golang':
            # golang does not have param, return markers
            return self.documentation.splitlines(keepends=True)
        else:
            raise NotImplementedError(
                'Documentation parsing for {0.language!r} in {0.docstyle!r}'
                ' has not been implemented yet'.format(self))

    def _parse_documentation_with_symbols(self,
                                          param_identifiers,
                                          exception_identifiers,
                                          return_identifiers):
        """
        Parses documentation based on parameter, exception and return symbols.

        :param param_identifiers:
            A tuple of two strings with which a parameter starts and ends.
        :param exception_identifiers:
            A tuple of two strings with which an exception starts and ends.
        :param return_identifiers:
            The string with which a return description starts.
        :return:
            The list of all the parsed sections of the documentation. Every
            section is a named tuple of either ``Description``, ``Parameter``,
            ``ExceptionValue`` or ``ReturnValue``.
        """
        lines = self.documentation.splitlines(keepends=True)

        parse_mode = self.Description

        cur_param = ''

        desc = ''
        parsed = []

        for line in lines:

            stripped_line = line.strip()

            if stripped_line.startswith(param_identifiers[0]):
                parse_mode = self.Parameter
                # param_offset contains the starting column of param's name.
                param_offset = line.find(
                    param_identifiers[0]) + len(param_identifiers[0])
                # splitted contains the whole line from the param's name,
                # which in turn is further divided into its name and desc.
                splitted = line[param_offset:].split(param_identifiers[1], 1)
                # parser breaks if param_identifiers[1] is not present.
                # This checks for space and then splits the line accordingly
                # to extract param's name and desc.
                if len(splitted) == 1:
                    splitted = line[param_offset:].split(' ', 1)
                cur_param = splitted[0].strip()

                param_desc = splitted[1]
                # parsed section is added to the final list.
                parsed.append(self.Parameter(name=cur_param, desc=param_desc))

            elif stripped_line.startswith(exception_identifiers[0]):
                parse_mode = self.ExceptionValue
                exception_offset = line.find(
                    exception_identifiers[0]) + len(exception_identifiers[0])
                splitted = line[exception_offset:].split(
                    exception_identifiers[1], 1)
                # parser breaks if exception_identifiers[1] is not present.
                # This checks for space and then splits the line accordingly
                # to extract exception's name and desc.
                if len(splitted) == 1:
                    splitted = line[exception_offset:].split(' ', 1)
                cur_exception = splitted[0].strip()

                exception_desc = splitted[1]
                parsed.append(self.ExceptionValue(
                    name=cur_exception, desc=exception_desc))

            elif stripped_line.startswith(return_identifiers):
                parse_mode = self.ReturnValue
                return_offset = line.find(
                    return_identifiers) + len(return_identifiers)
                retval_desc = line[return_offset:]
                parsed.append(self.ReturnValue(desc=retval_desc))

            # These conditions will take care if the parsed section
            # descriptions are not on the same line as that of it's
            # name. Further, adding the parsed section to the final list.
            elif parse_mode == self.ReturnValue:
                retval_desc += line
                parsed.pop()
                parsed.append(self.ReturnValue(desc=retval_desc))

            elif parse_mode == self.ExceptionValue:
                exception_desc += line
                parsed.pop()
                parsed.append(self.ExceptionValue(
                    name=cur_exception, desc=exception_desc))

            elif parse_mode == self.Parameter:
                param_desc += line
                parsed.pop()
                parsed.append(self.Parameter(name=cur_param, desc=param_desc))

            else:
                desc += line
                # This is inside a try-except for cases where the list
                # is empty and has nothing to pop.
                try:
                    parsed.pop()
                except IndexError:
                    pass
                parsed.append(self.Description(desc=desc))

        return parsed

    @classmethod
    def from_metadata(cls, doccomment, docstyle_definition,
                      marker, indent, position):
        r"""
        Assembles a list of parsed documentation comment metadata.

        This function just assembles the documentation comment
        itself, without the markers and indentation.

        >>> from coalib.bearlib.languages.documentation.DocumentationComment \
        ...     import DocumentationComment
        >>> from coalib.bearlib.languages.documentation.DocstyleDefinition \
        ...     import DocstyleDefinition
        >>> from coalib.results.TextPosition import TextPosition
        >>> Description = DocumentationComment.Description
        >>> Parameter = DocumentationComment.Parameter
        >>> python_default = DocstyleDefinition.load("python3", "default")
        >>> parsed_doc = [Description(desc='\nDescription\n'),
        ...               Parameter(name='age', desc=' Age\n')]
        >>> str(DocumentationComment.from_metadata(
        ...         parsed_doc, python_default,
        ...         python_default.markers[0], '    ',
        ...         TextPosition(0, 0)))
        '\nDescription\n:param age: Age\n'

        :param doccomment:
            The list of parsed documentation comment metadata.
        :param docstyle_definition:
            The ``DocstyleDefinition`` instance that defines what docstyle is
            being used in a documentation comment.
        :param marker:
            The markers to be used in the documentation comment.
        :param indent:
            The indentation to be used in the documentation comment.
        :param position:
            The starting position of the documentation comment.
        :return:
            A ``DocumentationComment`` instance of the assembled documentation.
        """
        assembled_doc = ''
        for section in doccomment:
            section_desc = section.desc.splitlines(keepends=True)

            if isinstance(section, cls.Parameter):
                assembled_doc += (docstyle_definition.metadata.param_start +
                                  section.name +
                                  docstyle_definition.metadata.param_end)

            elif isinstance(section, cls.ExceptionValue):
                assembled_doc += (docstyle_definition.metadata.exception_start
                                  + section.name
                                  + docstyle_definition.metadata.exception_end)

            elif isinstance(section, cls.ReturnValue):
                assembled_doc += docstyle_definition.metadata.return_sep

            assembled_doc += ''.join(section_desc)

        return DocumentationComment(assembled_doc, docstyle_definition, indent,
                                    marker, position)

    # we need to cache this function so as to construct full `self.range`
    @lru_cache(maxsize=1)
    def assemble(self):
        """
        Assembles parsed documentation to the original documentation.

        This function assembles the whole documentation comment, with the
        given markers and indentation.
        """
        lines = self.documentation.splitlines(keepends=True)
        assembled = self.indent + self.marker[0]
        if len(lines) == 0:
            return self.marker[0] + self.marker[2]
        assembled += lines[0]
        assembled += ''.join('\n' if line == '\n' and not self.marker[1]
                             else self.indent + self.marker[1] + line
                             for line in lines[1:])
        return (assembled if self.marker[1] == self.marker[2] else
                (assembled +
                 (self.indent if lines[-1][-1] == '\n' else '') +
                 self.marker[2]))
